#!/usr/bin/python3
# -*- coding: utf-8 -*-

# ------------------------------------------------------------------------------
# 
# Author: Gabriele Girelli
# Email: gigi.ga90@gmail.com
# Version: 0.0.1
# Description: Probe-design sub-section
# 
# ------------------------------------------------------------------------------



# DEPENDENCIES =================================================================

import logging

from ifpd.sections import app as rootApp
from ifpd.sections.probe_design.enquirer import Enquirer
from ifpd.sections.probe_design.queue import Queue
from ifpd.sections.probe_design.routes import Routes

# CLASSES ======================================================================

class App(rootApp.App):
	'''Probe Design Bottle App.

	Args:
		app_uri (string): section relative url.
		base_dir (string): section base directory.
		consumer (Enquirer): queue consumer.
		local_path (string): absolute path to app directory.
		qpath (string): absolute path to query folder.
		queue (Queue): query queue.
		root_path (string): webserver root absolute path.
		root_uri (string): root webserver url.
		route_list (string): probe_design.routes.Routes instance.
		section_path (string): sections relative path.
		tprefix (string): title prefix.
		vd (string): view data.
		vpath (string): absolute path to views folder.
		BUF_SIZE (int): queue size.
	'''
	
	vd = {}
	route_list = Routes()
	base_dir = 'probe_design'
	tprefix = 'Probe Designer ~ '
	vd = {}
	BUF_SIZE = 0
	MAX_CURR = 1

	def __init__(self, section_path, static_path,
		root_path, root_uri, app_uri,
		MAX_CURR = None, BUF_SIZE = None):
		'''Initialize.

		Args:
			section_path (string): sections relative path.
			root_path (string): webserver root absolute path.
			root_uri (string): root webserver url.
			app_uri (string): section relative url.
			MAX_CURR (int): maximum number of running tasks.
			BUF_SIZE (int): queue size, defaults to 0.
		'''
		
		# Run default initialization
		super(App, self).__init__()

		# Save input parameters
		self.section_path = section_path
		self.root_uri = root_uri
		self.app_uri = app_uri
		self.local_path = "%s%s" % (section_path, self.base_dir)
		self.static_path = static_path
		self.vpath = '%s/views/' % self.local_path
		self.qpath = '%s/query/' % self.static_path
		self.MAX_CURR = MAX_CURR
		self.BUF_SIZE = BUF_SIZE

		# Start setting up view dictionary
		self.vd['app_uri'] = "%s%s" % (root_uri, app_uri)
		self.vd['root_uri'] = root_uri
		self.vd['vpath'] = self.vpath
		self.vd['qpath'] = self.qpath
		self.vd['description'] = 'iFISH probe designer application.'
		self.vd['breadcrumbs'] = False
		self.vd['menu_template'] = None
		self.vd['SHOW_COOKIE_CONSENT_BANNER'] = False

		# Set default
		if not type(None) == type(MAX_CURR):
			self.MAX_CURR = MAX_CURR
		if not type(None) == type(BUF_SIZE):
			self.BUF_SIZE = BUF_SIZE

		# Logging config
		logging.basicConfig(level = logging.DEBUG,
			format = '(%(threadName)-9s) %(message)s')

		# Initialize queue
		self.queue = Queue(BUF_SIZE = self.BUF_SIZE, MAX_CURR = self.MAX_CURR)
		self.consumer = Enquirer(self.queue)
		self.consumer.start()

		# Save queue
		self.vd['queue'] = self.queue

		# Build routes
		self.build_routes()

		# Output
		return

# END ==========================================================================

################################################################################



# -*- coding: utf-8 -*-

'''
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
@description: biological file extension-specific methods.
'''



# ==============================================================================

import ifpd as fp
import io
import os
import pandas as pd
from tqdm import tqdm

class UCSCbed(object):
    '''Class to read a UCSC Bed format file into a pandas.DataFrame.'''

    FIELD_NAMES = ["chrom", "chromStart", "chromEnd",
        "name", "score", "strand", "thickStart", "thickEnd",
        "itemRgb", "blockCount", "blockSizes", "blockStarts"]
    FIELD_FORMAT = 'siisfsiisiss'
    FORMATTER = {'f':float, 'i':int, 's':str}

    def __init__(self, path, incrementChromEnd = False, bufferize = False):
        super(UCSCbed, self).__init__()
        self.path = path
        self.incrementChromEnd = incrementChromEnd
        self.nrecords = self.count_records()
        if not bufferize:
            self.__read()
    
    def count_records(self):
        return sum(1 for line in open(self.path))

    def __read(self):
        '''Reads the bed file into a pandas.DataFrame.'''
        assert os.path.isfile(self.path), f"bed file not found: '{self.path}'"

        bedDF = []
        with open(self.path, 'r+') as IH:
            header = next(IH)
            if header.startswith("browser") or header.startswith("track"):
                self.custom = True
                self.header = header.strip()
            else:
                bedDF.append(UCSCbed.parse_bed_line(header))
                self.custom = False
                self.header = None
            bedDF.extend([UCSCbed.parse_bed_line(line) for line in IH])

        self.df = pd.concat(bedDF)
        self.df.index = range(self.df.shape[0])
        self.ncols = self.df.shape[1]

    def buffer(self, parse = True, enforceBED3 = False):
        '''Reads the bed file and yields one line at a time. The content is not
        stored in the class. Each call to the buffer function moves forward to
        the next line. To restart, re-initialize the class with bufferize=True.
        Use enforceBED3 to strip any additional column. Use parse to get a
        formatted pd.DataFrame instead of a raw string.'''

        with open(self.path, 'r+') as IH:
            line = next(IH)
            if line.startswith("browser") or line.startswith("track"):
                self.custom = True
                self.header = line.strip()
            else:
                self.custom = False
                self.header = None

        with open(self.path, 'r+') as IH:
            if self.custom:
                next(IH)
            for line in tqdm(IH, total = self.nrecords-1):
                if parse:
                    yield UCSCbed.parse_bed_line(line,
                        enforceBED3, self.incrementChromEnd)
                else:
                    if self.incrementChromEnd:
                        line = line.strip().split("\t")
                        line[2] = str(int(line[2]) + 1)
                        line = "\t".join(line) + '\n'
                    yield line

    @staticmethod
    def parse_bed_line(line, enforceBED3 = False, incrementChromEnd = False):
        '''Parses and checks one line of a bed file.
        Does not work on header lines.'''
        
        line = line.strip().split("\t")
        assert 3 <= len(line), f"at least 3 fields required, {len(line)} found."
        line = line[:min(4 - enforceBED3, len(line))]

        lineDF = pd.DataFrame([
            UCSCbed.FORMATTER[UCSCbed.FIELD_FORMAT[i]](line[i])
            for i in range(len(line))]).transpose()
        lineDF.columns = UCSCbed.FIELD_NAMES[:len(line)]

        lineDF.iloc[0, 2] += incrementChromEnd

        return(lineDF)

    @staticmethod
    def add_sequence_to_raw_record(bedRecords):
        for record in bedRecords:
            assert type("") == type(record)
            record = record.strip().split("\t")

            assert_msg = "each record must have at least 3 columns,"
            assert_msg += f" {len(record)} found."
            assert 3 <= len(record), assert_msg

            if 4 > len(record):
                oligoSequence = fp.web.get_sequence_from_UCSC(
                    (record[0], int(record[1]), int(record[2])), args.refGenome)
                record.append(oligoSequence)
            record = record[:4]
            yield("\t".join(record) + '\n')

    @staticmethod
    def add_sequence_to_parsed_record(bedRecords):
        for record in bedRecords:
            assert type(pd.DataFrame()) == type(bedRecords)

            assert_msg = "each record must have at least 3 columns,"
            assert_msg += f" {record.shape[0]} found."
            assert 3 <= record.shape[0]
            
            if 4 > record.shape[1]:
                region = record.loc[:, ['chrom', 'chromStart', 'chromEnd']
                    ].iloc[0, :].tolist()
                oligoSequence = fp.web.get_sequence_from_UCSC(
                    tuple(region), args.refGenome)
                record.append(oligoSequence)
            record = record.iloc[0, :4]
            yield(record)

    def isBEDN(self, n):
        n = min(n, 12)
        return self.ncols == n

    def getBEDN(self, n):
        n = min(n, 12)
        assert self.ncols >= n
        return self.df.iloc[:, :n]

    def mkBEDN(self, n):
        assert n <= 12
        if n < self.ncols:
            self.df = self.df.iloc[:, :n]
            self.ncols = n

# END ==========================================================================

################################################################################



# -*- coding: utf-8 -*-

'''
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
@description: methods related to database query.
'''



# ==============================================================================

import matplotlib
matplotlib.use('svg')
import matplotlib.pyplot as plt
import matplotlib.patches as patches

import configparser
from ggc.args import check_threads
import ifpd as fp
from joblib import Parallel, delayed
import numpy as np
import os
import pandas as pd
from tqdm import tqdm

class OligoDatabase(object):
    """FISH-ProDe Oligonucleotide Database class."""

    def __init__(self, dbDirPath, hasNetwork = True,
        UCSC_DAS_URI = fp.web.UCSC_DAS_URI):
        super(OligoDatabase, self).__init__()
        self.dirPath = dbDirPath
        self.hasNetwork = hasNetwork
        self.UCSC_DAS_URI = UCSC_DAS_URI
        self.chromData = {}

        assert_msg = f'expected folder, file found: "{self.dirPath}"'
        assert not os.path.isfile(self.dirPath), assert_msg

        assert_msg = f'database folder not found: "{self.dirPath}"'
        assert os.path.isdir(self.dirPath), assert_msg

        configPath = os.path.join(self.dirPath, ".config")
        assert_msg = f'missing "{configPath}" file.'
        assert os.path.isfile(configPath), assert_msg

        with open(configPath, 'r') as IH:
            self.config = configparser.ConfigParser()
            self.config.read_string("".join(IH.readlines()))

        if self.hasNetwork:
            refGenome = self.get_reference_genome()
            refGenomeChecked = fp.web.check_reference_genome(
                refGenome, self.UCSC_DAS_URI)
            assert refGenomeChecked, f'genome "{refGenome}" not found @UCSC'

    def check_overlaps(self):
        hasOverlaps = False
        for chrom, chromData in self.chromData.items():
            startPositions = np.array(chromData.iloc[:-1, 0])
            endPositions = np.array(chromData.iloc[1:, 1]) - 1
            foundOverlaps = any(startPositions <= endPositions)
            hasOverlaps |= foundOverlaps
        return hasOverlaps == self.has_overlaps()

    def get_oligo_length_range(self):
        '''Reads oligo length range from Database .config'''
        return (
            self.config.getint('OLIGOS', 'min_length'),
            self.config.getint('OLIGOS', 'max_length')
        )

    def get_oligo_min_dist(self):
        '''Reads consecutive oligo minimum distance from Database .config'''
        return self.config.getint('OLIGOS', 'min_dist')

    def get_name(self):
        '''Reads Database name from Database .config'''
        return self.config['DATABASE']['name']

    def get_reference_genome(self):
        '''Reads reference genome from Database .config'''
        return self.config['DATABASE']['refGenome']

    def has_overlaps(self):
        '''Reads overlaps status from Database .config'''
        return self.config.getboolean('OLIGOS', 'overlaps')

    def has_sequences(self):
        '''Reads sequence status from Database .config'''
        return self.config.getboolean('OLIGOS', 'sequence')

    def has_chromosome(self, chrom):
        return chrom in os.listdir(self.dirPath)

    def read_chromosome(self, chrom):
        assert self.has_chromosome(chrom)
        
        chromPath = os.path.join(self.dirPath, chrom)
        chromData = pd.read_csv(chromPath, '\t', header = None)
        chromData.columns = fp.bioext.UCSCbed.FIELD_NAMES[
            1:(chromData.shape[1]+1)]

        assert_msg = f'found empty chromosome file: "{chromPath}"'
        assert 0 != chromData.shape[0], assert_msg

        assert_msg = f'missing columns in "{chromPath}"'
        assert 2 <= chromData.shape[1], assert_msg

        assert_msg = f'found unsorted file: "{chromPath}"'
        assert all(chromData.iloc[:, 0].diff()[1:] > 0), assert_msg

        assert_msg = f'found unsorted file: "{chromPath}"'
        assert all(chromData.iloc[:, 1].diff()[1:] >= 0), assert_msg

        oligoMinDist = self.get_oligo_min_dist()
        startValues = chromData.iloc[1:, 0].values
        endValues = chromData.iloc[:-1, 1].values
        oligoMinD_observed = min(startValues - endValues)
        assert_msg = 'oligo min distance does not match: '
        assert_msg += f'{oligoMinD_observed} instead of {oligoMinDist}.'
        assert oligoMinD_observed == oligoMinDist, assert_msg

        oligoLengthRange = self.get_oligo_length_range()
        oligoLengthList = chromData.iloc[:, 1] - chromData.iloc[:, 0]
        assert_msg = f'oligo too small for ".config" in "{chromPath}"'
        assert all(oligoLengthList >= oligoLengthRange[0]), assert_msg
        assert_msg = f'oligo too big for ".config" in "{chromPath}"'
        assert all(oligoLengthList <= oligoLengthRange[1]), assert_msg

        if self.has_overlaps():
            assert_msg = f'overlaps status mismatch in "{chromPath}"'
            assert self.check_overlaps(), assert_msg

        if self.has_sequences():
            assert_msg = f'missing sequence columns in "{chromPath}"'
            assert chromData.shape[1] >= 3, assert_msg

        if self.hasNetwork and self.has_sequences():
            for i in range(chromData.shape[0]):
                chromStart, chromEnd, sequence = tuple(
                    chromData.iloc[i, :3].tolist())

                assert_msg = 'sequence length does not match oligo'
                assert_msg += f' (#{i}) length'
                assert len(sequence) == chromEnd-chromStart, assert_msg

                assert_status, msg = fp.web.check_sequence(
                    (chrom, chromStart, chromEnd), sequence,
                    self.get_reference_genome(),
                    UCSC_DAS_URI = self.UCSC_DAS_URI)
                assert_msg = f'sequence of oligo #{i} does not match'
                assert_msg += ' UCSC' + msg
                assert assert_status, assert_msg
        
        if self.hasNetwork:
            chromSizeCheck = fp.web.check_chromosome_size(
                chrom, chromData.iloc[-1, 1],
                self.get_reference_genome(),
                UCSC_DAS_URI = self.UCSC_DAS_URI)
            assert_msg = f'chromosome size not respected: "{chromPath}"'
            assert chromSizeCheck, assert_msg

        self.chromData[chrom] = chromData

    def read_all_chromosomes(self, verbose):
        chromList = [d for d in os.listdir(self.dirPath)
            if not d.startswith('.')]
        assert 0 < len(chromList), "no chromosome files found in {self.dirPath}"
        chromList = tqdm(chromList) if verbose else chromList

        for chrom in chromList:
            self.read_chromosome(chrom)

class OligoProbe(object):
    """Class for probe management."""

    def __init__(self, chrom, oligos, database):
        super(OligoProbe, self).__init__()
        self.chrom = chrom
        self.oligoData = oligos
        self.refGenome = database.get_reference_genome()
        self.chromStart = self.oligoData.iloc[:, 0].min()
        self.chromEnd = self.oligoData.iloc[:, 1].max()
        self.midpoint = (self.chromStart + self.chromEnd) / 2
        self.size = self.chromEnd - self.chromStart
        self.homogeneity = self.get_probe_homogeneity()

    def __str__(self):
        s  = f'[{self.refGenome}]'
        s += f'{self.chrom}:{self.chromStart}-{self.chromEnd};'
        s += f' oligoSpread: {self.homogeneity}'
        return s

    def asDataFrame(self, region = None):
        if type(None) == type(region):
            return pd.DataFrame.from_dict({
                'chrom' : [self.chrom],
                'chromStart' : [self.chromStart],
                'chromEnd' : [self.chromEnd],
                'refGenome' : [self.refGenome],
                'midpoint' : [self.midpoint],
                'size' : [self.size],
                'homogeneity' : [self.homogeneity]
            })
        else:
            return pd.DataFrame.from_dict({
                'chrom' : [self.chrom],
                'chromStart' : [self.chromStart],
                'chromEnd' : [self.chromEnd],
                'refGenome' : [self.refGenome],
                'midpoint' : [self.midpoint],
                'size' : [self.size],
                'homogeneity' : [self.homogeneity],
                'regChromStart' : region[0],
                'regChromEnd' : region[1]
            })

    def get_probe_centrality(self, region):
        '''Calculate centrality, as location of the probe midpoint relative to
        the region midpoint. 1 when overlapping, 0 when the probe midpoint is
        at either of the region extremities.'''
        region_halfWidth = (region[2] - region[1]) / 2
        region_midPoint = region[1] + region_halfWidth
        centrality = (region_halfWidth - abs(
            region_midPoint - self.midpoint)) / region_halfWidth
        return centrality

    def get_probe_size(self):
        '''Probe size, defined as difference between start of first oligo
        and end of the last one.'''
        return self.oligoData.iloc[:, 1].max() - self.oligoData.iloc[:, 0].min()

    def get_probe_homogeneity(self):
        '''Probe homogeneity, as the inverse of the standard deviation of the
        distance between consecutive oligos, calculated from their start
        position, disregarding oligo length.'''
        std = np.std(np.diff(self.oligoData.iloc[:, 1]))
        return np.inf if 0 == std else 1/std

    def describe(self, region, path = None):
        '''Builds a small pd.DataFrame describing a probe.'''
        description = pd.DataFrame.from_dict({
            'chrom' : [region[0]],
            'chromStart' : [self.oligoData.iloc[:, 0].min()],
            'chromEnd' : [self.oligoData.iloc[:, 1].max()],
            'centrality' : [self.get_probe_centrality(region)],
            'size' : [self.size],
            'homogeneity' : [self.homogeneity]
        })

        if not type(None) == type(path):
            config = configparser.ConfigParser()
            config['REGION'] = {
                'chrom' : region[0],
                'chromStart' : region[1],
                'chromEnd' : region[2]
            }
            config['PROBE'] = {
                'chrom' : description['chrom'].values[0],
                'chromStart' : description['chromStart'].values[0],
                'chromEnd' : description['chromEnd'].values[0],
                'nOligo' : self.oligoData.shape[0]
            }
            config['FEATURES'] = {
                'centrality' : description['centrality'].values[0],
                'size' : description['size'].values[0],
                'homogeneity' : description['homogeneity'].values[0]
            }
            with open(path, 'w+') as OH:
                config.write(OH)

        return description

    def get_fasta(self, path = None, prefix = ""):
        if not prefix.startswith(" "):
            prefix = " " + prefix
            
        fasta = ""
        for i in self.oligoData.index:
            oligo = self.oligoData.loc[i, :]
            chromStart, chromEnd, sequence = oligo[:3]
            fasta += f'>{prefix}oligo_{i} [{self.refGenome}]'
            fasta += f'{self.chrom}:{chromStart}-{chromEnd}\n'
            fasta += f'{sequence}\n'

        if not type(None) == type(path):
            assert os.path.isdir(os.path.dirname(path))
            with open(path, 'w+') as OH:
                OH.write(fasta)

        return fasta

    def get_bed(self, path = None, prefix = ""):
        if not prefix.endswith("_"):
            prefix += "_"

        bed = f'track description="ref:{self.refGenome}"\n'
        for i in self.oligoData.index:
            oligo = self.oligoData.loc[i, :]
            chromStart, chromEnd, sequence = oligo[:3]
            bed += f'{self.chrom}\t{chromStart}\t{chromEnd}\t'
            bed += f'{prefix}oligo_{i}\n'

        if not type(None) == type(path):
            assert os.path.isdir(os.path.dirname(path))
            with open(path, 'w+') as OH:
                OH.write(bed)

        return bed

    def _plot_region(self, outputDir, region):
        fig = plt.figure()

        chrom, start, stop = region
        plt.plot([start, stop], [0, 0], 'k', linewidth = 4.0, label = 'Genome')
        plt.plot([start + (stop-start)/2., start + (stop-start)/2.],
            [-1, 1], 'r--', label = 'Window center')

        plt.plot([self.chromStart, self.chromEnd], [0, 0],
            'c-', linewidth = 4.0, label = 'Probe')
        plt.plot([self.chromStart+self.size/2., self.chromStart+self.size/2.],
            [-1, 1], 'c--',label = 'Probe center')

        plt.gca().axes.get_yaxis().set_visible(False)
        plt.suptitle('%s:%d-%d' % (chrom, start, stop))
        plt.xlabel('genomic coordinate [nt]')
        plt.legend(fontsize = 'small', loc = 'best')

        plt.savefig(os.path.join(outputDir, 'window.png'),
            format = 'png', bbox_inches = 'tight')
        plt.close(fig)

    def _plot_oligo(self, outputDir):
        fig = plt.figure(figsize = (20, 5))
        genome_handle, = plt.plot([self.chromStart, self.chromEnd],
            [0, 0], 'k', linewidth = 4.0, label = 'Genome')

        for i in self.oligoData.index:
            oligo = self.oligoData.loc[i, :]
            oligo_midpoint = (oligo['chromStart'] + oligo['chromStart']) / 2.
            oligo_handle, = plt.plot([oligo['chromStart'], oligo['chromEnd']],
                [0, 0], 'c', linewidth = 2.0, label = 'Oligo')
            oligoCenter_handle, = plt.plot(
                [oligo_midpoint, oligo_midpoint],
                [-.1, .1], 'c:', label = 'Oligo center')

        plt.gca().axes.get_yaxis().set_visible(False)
        plt.ylim((-.5, .5))

        plt.gca().axes.get_xaxis().set_ticks(list(range(
            self.chromStart, self.chromEnd,
            int((self.chromEnd-self.chromStart)/5.))))

        plt.legend(handles = [genome_handle, oligo_handle, oligoCenter_handle],
            fontsize = 'small', loc = 'best')
        plt.suptitle((f'{self.chrom}:{self.chromStart}-{self.chromEnd}'))
        plt.xlabel('genomic coordinate [nt]')

        plt.savefig(os.path.join(outputDir, 'probe.png'),
            format = 'png', bbox_inches = 'tight')
        plt.close(fig)

    def _plot_oligo_distr(self, outputDir):
        fig = plt.figure()

        plt.plot(
            [0, self.oligoData.shape[0]-1],
            [self.chromStart, self.chromEnd], 'k-',
            label = 'Homogeneous distribution')
        plt.plot(
            list(range(self.oligoData.shape[0])),
            self.oligoData['chromStart'].values,
            'r.', label = 'Oligo')

        plt.suptitle(f'{self.chrom}:{self.chromStart}-{self.chromEnd}')
        plt.xlabel('oligo number')
        plt.ylabel('genomic coordinate [nt]')
        plt.legend(fontsize = 'small', loc = 'best')

        plt.savefig(os.path.join(outputDir, 'oligo.png'), format = 'png')
        plt.close(fig)

    def _plot_oligo_distance(self, outputDir):
        fig = plt.figure()

        startPositions = self.oligoData.iloc[1:, 0].values
        endPositions = self.oligoData.iloc[:-1, 1].values
        diffs = startPositions - endPositions
        plt.hist(diffs, density = 1, facecolor = 'green', alpha = .5)
        density = fp.stats.calc_density(diffs, alpha = .5)
        plt.plot(density['x'].tolist(), density['y'].tolist(), 'b--',
            label = 'Density distribution')

        plt.suptitle(f'{self.chrom}:{self.chromStart}-{self.chromEnd}')
        plt.xlabel('Distance between consecutive oligos [nt]')
        plt.ylabel('Density')
        plt.legend(fontsize = 'small', loc = 'best')

        plt.savefig(os.path.join(outputDir, 'distance.png'), format = 'png')
        plt.close(fig)

    def plot(self, outputDir, region):
        assert os.path.isdir(outputDir), f'folder not found: "{outputDir}"'
        self._plot_region(outputDir, region)
        self._plot_oligo(outputDir)
        self._plot_oligo_distr(outputDir)
        self._plot_oligo_distance(outputDir)

def describe_candidate(candidate, queried_region):
    return candidate.describe(queried_region)

class ProbeFeatureTable(object):
    FEATURE_SORT = {
        'centrality' : {'ascending':False},
        'size' : {'ascending':True},
        'homogeneity' : {'ascending':False},
    }

    def __init__(self, candidateList, queried_region,
        verbose = False, threads = 1):
        super(ProbeFeatureTable, self).__init__()
        assert 0 < len(candidateList)
        
        self.data = []
        if 1 != threads:
            verbose = 1 if verbose else 0
            self.data = Parallel(n_jobs = threads,
                backend = 'threading', verbose = verbose)(
                delayed(describe_candidate)(candidate, queried_region)
                for candidate in candidateList)
        else:
            candidateList = tqdm(candidateList) if verbose else candidateList
            for candidate in candidateList:
                self.data.append(candidate.describe(queried_region))
        self.data = pd.concat(self.data)
        self.data.index = range(self.data.shape[0])

        self.discarded = None

    def reset(self):
        self.data = pd.concat([self.discarded, self.data])
        self.discarded = None
        self.data.sort_index(inplace = True)

    def keep(self, condition, cumulative = False):
        if not cumulative:
            self.reset()
        self.discarded = pd.concat([self.discarded,
            self.data.loc[np.logical_not(condition), :]])
        self.data = self.data.loc[condition, :]

    def filter(self, feature, thr, cumulative = False):
        assert_msg = f'fetature "{feature}" not recognized.'
        assert feature in self.FEATURE_SORT.keys(), assert_msg

        if not cumulative:
            self.reset()

        self.rank(feature)

        best_feature = self.data[feature].values[0]
        feature_delta = best_feature * thr
        feature_range = (best_feature-feature_delta, best_feature+feature_delta)

        discardCondition = [self.data[feature] < feature_range[0]]
        discardCondition.append(self.data[feature] > feature_range[1])
        discardCondition = np.logical_or(*discardCondition)
        self.discarded = pd.concat([self.discarded,
            self.data.loc[discardCondition, :]])
        self.data = self.data.loc[np.logical_not(discardCondition), :]

        return (feature_range, feature)

    def rank(self, feature):
        self.data = self.data.sort_values(feature,
            ascending = self.FEATURE_SORT[feature]['ascending'])

class GenomicWindow(object):
    def __init__(self, chrom, start, size):
        super(GenomicWindow, self).__init__()
        self.chrom = chrom
        self.chromStart = start
        self.chromEnd = start + size
        self.midpoint = (self.chromStart + self.chromEnd) / 2
        self.size = size
        self.probe = None

    def __str__(self):
        s  = f'[GenomicWindow]{self.chrom}:{self.chromStart}-{self.chromEnd}\n'
        s += f' Probe: {self.probe}'
        return s

    def asRegion(self):
        return (self.chrom, self.chromStart, self.chromEnd)

    def has_probe(self):
        return type(None) != type(self.probe)

    def shift(self, n):
        return GenomicWindow(self.chrom, self.chromStart + n, self.size)

class GenomicWindowList(object):
    '''Both a list genomic window and associated probe set.'''

    data = []

    def __init__(self, windows = None):
        super(GenomicWindowList, self).__init__()
        if type(None) != type(windows):
            self.data = windows

    def __getitem__(self, i):
        return self.data[i]

    def __iter__(self):
        for window in self.data:
            yield window

    def __len__(self):
        return len(self.data)

    def add(self, chrom, start, size):
        self.data.append(GenomicWindow(chrom, start, size))

    def count_probes(self):
        return sum([w.has_probe() for w in self])

    def shift(self, n):
        return GenomicWindowList([window.shift(n) for window in self.data])

    def sort(self):
        midpointList = np.array([w.midpoint for w in self.data])
        self.data = [self.data[i] for i in np.argsort(midpointList)]

    def calc_probe_size_and_homogeneity(self):
        if 3 > self.count_probes():
            return np.nan

        probeData = pd.concat([w.probe.asDataFrame((w.chromStart, w.chromEnd))
            for w in self if type(None) != type(w.probe)])
        size_std = probeData['size'].values.std()
        probe_homogeneity = np.std(probeData['chromStart'].values[1:] - 
            probeData['chromEnd'].values[:-1])
        return 2 / (size_std + probe_homogeneity)

    def asDataFrame(self):
        starts = []
        ends = []
        sizes = []
        for window in self:
            starts.append(window.chromStart)
            ends.append(window.chromEnd)
            sizes.append(window.size)
        return pd.DataFrame.from_dict({
            'chromStart' : starts,
            'chromEnd' : ends,
            'size' : sizes
        })

    def _plot_probe_set(self, outputDir, region):
        fig = plt.figure(figsize = (20, 5))

        for wi in range(len(self)):
            window = self[wi]
            plt.axvline(window.chromStart, ymin = -1, ymax = 1, color = 'k')
            if not window.has_probe():
                plt.gca().add_patch(patches.Rectangle(
                    (window.chromStart, -1), window.size, 2, color = 'r'))
            plt.gca().text(window.midpoint, .5, wi + 1)
        plt.axvline(self[0].chromStart, ymin = -1, ymax = 1,
            color = 'k', linestyle = ':')
        plt.axvline(self[-1].chromEnd, ymin = -1, ymax = 1,
            color = 'k', linestyle = ':')

        plt.gca().add_patch(patches.Rectangle(
            (region[1], -1), self[0].chromStart - region[1], 2, color = 'k'))
        plt.gca().add_patch(patches.Rectangle((self[-1].chromEnd, -1),
            region[2]-self[-1].chromEnd, 2, color = 'k'))

        genome, = plt.plot([region[1], region[2]], [0, 0], 'k',
            linewidth = 4.0, label = 'Genome')

        for probe in [w.probe for w in self if type(None) != type(w.probe)]:
            oligo, = plt.plot([probe.chromStart, probe.chromEnd], [0, 0], 'c',
                linewidth = 2.0, label = 'Probe')
            oligo_center, = plt.plot([probe.midpoint, probe.midpoint],
                [-.1, .1], 'c:', label = 'Probe center', linewidth = 2.0)

        plt.gca().axes.get_yaxis().set_visible(False)
        plt.ylim((-1,1))
        msg = 'Empty windows are reported in red.'
        plt.suptitle(f'Region: {region[0]}:{region[1]}-{region[2]}' +
            ' & Probe set: ' +
            f'{self[0].chrom}:{self[0].chromStart}-{self[-1].chromEnd}\n' + msg)
        plt.xlabel('genomic coordinate [nt]')

        plt.savefig(os.path.join(outputDir, 'windows.png'),
            format = 'png', bbox_inches = 'tight')
        plt.close(fig)

    def _plot_probe_distr(self, outputDir):
        fig = plt.figure()

        probes = [w.probe for w in self if type(None) != type(w.probe)]

        plt.plot([0, len(probes) - 1],
            [probes[0].midpoint, probes[-1].midpoint], 'k-',
            label = 'Homogeneous distribution')
        plt.plot(list(range(len(probes))), [p.midpoint for p in probes],
            'r.', label = 'Probe')

        plt.suptitle(f'{probes[0].chrom}:' +
            f'{probes[0].chromStart}{probes[-1].chromEnd}')
        plt.xlabel('probe number')
        plt.ylabel('genomic coordinate [nt]')
        plt.xlim((-1, len(probes)))
        plt.legend(fontsize = 'small', loc = 'best')

        plt.savefig(os.path.join(outputDir, 'distr.png'),
            format = 'png', bbox_inches = 'tight')
        plt.close(fig)

    def _plot_probe_distance(self, outputDir, region):
        fig = plt.figure()

        probes = [w.probe for w in self if type(None) != type(w.probe)]

        starts = np.array([p.chromStart for p in probes][1:])
        ends = np.array([p.chromEnd for p in probes][:-1])
        diffs = starts - ends
        plt.hist(diffs, density = 1, facecolor = 'green', alpha = .5)

        density = fp.stats.calc_density(diffs, alpha = .5)
        plt.plot(density['x'].tolist(), density['y'].tolist(), 'b--',
            label = 'Density distribution')

        plt.suptitle(f'{probes[0].chrom}:' +
            f'{probes[0].chromStart}{probes[-1].chromEnd}')
        plt.xlabel('Distance between consecutive probes [nt]')
        plt.ylabel('Density')
        plt.legend(fontsize = 'small', loc = 'best')

        plt.savefig(os.path.join(outputDir, 'distance.png'),
            format = 'png', bbox_inches = 'tight')
        plt.close(fig)

    def export(self, path, region):
        probe_counter = 0
        fasta = ""
        bed = ""
        for window in self:
            if window.has_probe():
                probe_dirName = f'probe_{probe_counter}'
                probe_dirPath = os.path.join(path, probe_dirName)
                probe = window.probe

                assert not os.path.isfile(probe_dirPath)
                assert not os.path.isdir(probe_dirPath)
                os.mkdir(probe_dirPath)

                probe.describe(window.asRegion(), os.path.join(
                    probe_dirPath, f'probe_{probe_counter}.config'))
                fasta += probe.get_fasta(os.path.join(
                    probe_dirPath, f'probe_{probe_counter}.fasta'),
                    f'probe{probe_counter}')
                probe_bed = probe.get_bed(os.path.join(
                    probe_dirPath, f'probe_{probe_counter}.bed'),
                    f'probe{probe_counter}')
                bed += "\n".join(probe_bed.split("\n")[1:])
                probe.plot(probe_dirPath, window.asRegion())

                probe_counter += 1

        self._plot_probe_set(path, region)
        self._plot_probe_distr(path)
        self._plot_probe_distance(path, region)

        return (fasta, bed)

# END ==========================================================================

################################################################################



# -*- coding: utf-8 -*-

'''
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
@description: statistical methods.
'''

# ==============================================================================

import numpy as np
from scipy import stats as spStats

def calc_density(data, **kwargs):
    '''Calculate the Gaussian KDE of the provided data series.'''
    sigma = .2 if not 'sigma' in list(kwargs.keys()) else kwargs['sigma']
    nbins = 1000 if not 'nbins' in list(kwargs.keys()) else kwargs['nbins']

    # If only one nucleus was found
    if 1 == len(data):
        f = eval(f'lambda x: 1 if x == {data[0]} else 0')
        f = np.vectorize(f)
        return({
            'x' : np.array([data[0]]),
            'y' : np.array([1]), 'f' : f
        })

    data = data[np.logical_not(np.isnan(data))]
    density = spStats.gaussian_kde(data)
    density.covariance_factor = lambda : sigma
    density._compute_covariance()
    return({
        'x' : np.linspace(min(data), max(data), nbins),
        'y' : density(np.linspace(min(data), max(data), nbins)),
        'f' : density
    })


# END ==========================================================================

################################################################################



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# ------------------------------------------------------------------------------
# 
# Author: Gabriele Girelli
# Email: gigi.ga90@gmail.com
# Version: 0.0.1
# Description: Probe-design Route manager
# 
# ------------------------------------------------------------------------------



# DEPENDENCIES =================================================================

import logging
import bottle as bot

# CLASSES ======================================================================

class Routes:
	'''Routes class.'''

	# Empty routes dictionary
	data = {}

	def __init__(self):
		'''
		Create new routes here using the add_route method.
		'''

		# Static files ---------------------------------------------------------
		
		route = '/<dname:re:(css|js|fonts|images|documents)>/<path>'
		self.add_route('static_file', 'route', route)

		route = '/documents/<path>/mimetype/<mt1>/<mt2>'
		self.add_route('static_raw_file_download', 'route', route)

		# Pages ----------------------------------------------------------------

		self.add_route('home', 'route', '/')
		self.add_route('home', 'view', 'home.tpl')

		# Errors ---------------------------------------------------------------

		self.add_route('error404', 'error', 404)
		self.add_route('error500', 'error', 500)

		return

	def add_route(self, k, t, v):
		'''Create a new route or add data to an existing one.

		Args:
			k (string): route name, identical to the route function.
			t (string): the data type (route|view|get|post|error).
			v (string|int): the data value.
		'''

		# Create a new route if it does not exist
		if not k in self.data.keys():
			self.data[k] = {}

		# Update existing route
		self.data[k][t] = v

		# Output
		return

	# Pages --------------------------------------------------------------------

	def home(routes, self):
		'''Default home-page.

		Args:
			self (App): ProbeDesigner.App instance.
		'''
		d = self.vd.copy()
		d['title'] = 'Default Home-page'
		d['description'] = 'Nothing to say...'
		return(d)

	# Static files -------------------------------------------------------------

	def static_file(routes, self, dname, path):
		'''Access static files.

		Args:
			self (App): ProbeDesigner.App instance.
			dname (string): file type.
			path (string): file name.
		'''
		return(bot.static_file(path, '%s/%s/' % (self.local_path, dname)))

	def static_raw_file_download(routes, self,
		path, mt1, mt2):
		'''Download raw static files.

		Args:
			self (App): ProbeDesigner.App instance.
			path (string): file name.
			mt1 (string): first part of the mimetype.
			mt2 (string): second part of the mimetype.
		'''
		ipath = '%s/documents/' % self.local_path
		mt = '%s/%s' % (mt1, mt2)
		return(bot.static_file(path, ipath, mimetype = mt))

	# Error --------------------------------------------------------------------

	def error404(routes, self, error):
		'''Error 404.

		Args:
			self (App): ProbeDesigner.App instance.
			error: error data.
		'''
		return('ERROR 404: nothing here, sorry :(')

	def error500(routes, self, error):
		'''Error 500.

		Args:
			self (App): ProbeDesigner.App instance.
			error: error data.
		'''
		return('ERROR 500: nothing here, sorry :(')

# END ==========================================================================

################################################################################



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# ------------------------------------------------------------------------------
# 
# Author: Gabriele Girelli
# Email: gigi.ga90@gmail.com
# Version: 0.0.1
# Description: Probe-design Queue consumer (database enquirer)
# 
# ------------------------------------------------------------------------------



# DEPENDENCIES =================================================================

import configparser
import datetime
import logging
import os
import subprocess as sp
import threading
import time

# CLASSES ======================================================================

class Enquirer(threading.Thread):
	'''Database enquirer.'''

	def __init__(self, queue):
		'''Instance method.

		Args:
			queue (Queue): Queue object which will contain the queries.
		'''

		# Initializer
		super(Enquirer, self).__init__()

		# Set class vars
		self.queue = queue

		# Close
		return

	def readQueryConfig(self, queryDir):
		config = configparser.ConfigParser()
		with open(f'{queryDir}.config', 'r') as IH:
			config.read_string("".join(IH.readlines()))
		return config

	def writeQueryConfig(self, queryDir, config):
		with open(f'{queryDir}.config', 'w+') as OH:
			config.write(OH)

	def run(self):
		'''Run one query from the queue.'''

		while True:
			if not self.queue.empty():
				cmd = self.queue.get()

				# If the queue released a task
				if not type(None) == type(cmd):
					if 'ifpd_query_set' == cmd[0]:
						outdir_id = 4
					elif 'ifpd_query_probe' == cmd[0]:
						outdir_id = 3

					query_id = os.path.basename(cmd[outdir_id])


					logging.debug(f'Running query "{query_id}"')
					timestamp = time.time()
					isotimestamp = datetime.datetime.fromtimestamp(
						timestamp).isoformat()
					config = self.readQueryConfig(cmd[outdir_id])
					config['GENERAL']['status'] = 'running'
					config['WHEN']['start_time'] = f'{timestamp}'
					config['WHEN']['start_isotime'] = isotimestamp
					self.writeQueryConfig(cmd[outdir_id], config)

					sp.call(cmd)
					timestamp = time.time()
					isotimestamp = datetime.datetime.fromtimestamp(
						timestamp).isoformat()
					logging.debug(f'Finished query "{query_id}"')
					config = self.readQueryConfig(cmd[outdir_id])
					config['GENERAL']['status'] = 'done'
					config['WHEN']['done_time'] = f'{timestamp}'
					config['WHEN']['done_isotime'] = isotimestamp
					self.writeQueryConfig(cmd[outdir_id], config)
					cmd = self.queue.task_done(cmd)

		return

# END ==========================================================================

################################################################################



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# ------------------------------------------------------------------------------
# 
# Author: Gabriele Girelli
# Email: gigi.ga90@gmail.com
# Version: 0.0.1
# Description: Probe-design queriesQueue
# 
# ------------------------------------------------------------------------------



# DEPENDENCIES =================================================================

import logging
import queue as q
import subprocess as sp

# CLASSES ======================================================================

class Queue(q.Queue):
	'''Database query Queue.

	Args:
		MAX_CURR (int): maximum number of simultaneously released tasks.
		doing (list): list of currently released tasks (i.e., running).
		done (list): list of completed taks.
	'''

	MAX_CURR = 1
	doing = []
	done = []

	def __init__(self, MAX_CURR = None, **kwargs):
		'''Instance method.

		Args:
			n_doing (int): max number of simultaneously running tasks.
		'''
		if not type(None) == type(MAX_CURR):
			self.MAX_CURR = MAX_CURR
		super(Queue, self).__init__()
		return

	def get(self, **kwargs):
		'''Extend original get method by setting up doing.
		Also forces only one element to be running at a time.
		'''

		# Stop if already running
		if not self.MAX_CURR >= len(self.doing):
			return

		# Call original method
		released = super(Queue, self).get(**kwargs)

		# Append released element to released list
		self.doing.append(released)

		# Output released element
		return(released)

	def task_done(self, done, **kwargs):
		'''Extend original task_done method by adding doing and done features.'''

		# If unknown task, kill execution
		if not done in self.doing:
			return

		# Set the task as completed
		self.done.append(self.doing.pop(self.doing.index(done)))

		# Call original method
		super(Queue, self).task_done(**kwargs)

		# Stop
		return

# END ==========================================================================

################################################################################



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# ------------------------------------------------------------------------------
# 
# Author: Gabriele Girelli
# Email: gigi.ga90@gmail.com
# Version: 0.0.1
# Description: Oligopool-design sub-section
# 
# ------------------------------------------------------------------------------



# DEPENDENCIES =================================================================

import bottle as bot

# CLASSES ======================================================================

class App(bot.Bottle):
	'''WebServer section App.'''

	def __init__(self):
		'''Initialize.'''

		# Run default initialization
		super(App, self).__init__()

		# Output
		return

	def route_builder(self, page, data):
		'''Generate routes based on route data.

		Args:
			page (string): route name, identical to route function name.
			data (dict): rout data.
		'''

		# If a proper route is requested
		if 'route' in data.keys():
			
			# Load template if requested
			if 'view' in data.keys():
				@self.route(data['route'])
				@bot.view(self.vpath + data['view'])
				def route(**kwargs):
					return(getattr(self.route_list, page)(self, **kwargs))
			else:
				@self.route(data['route'])
				def route(**kwargs):
					return(getattr(self.route_list, page)(self, **kwargs))

		# Get reception
		elif 'get' in data.keys():
			@self.get(data['get'])
			def route(**kwargs):
				return(getattr(self.route_list, page)(self, **kwargs))

		# Post reception
		elif 'post' in data.keys():
			@self.post(data['post'])
			def route(**kwargs):
				return(getattr(self.route_list, page)(self, **kwargs))

		# Error pages
		elif 'error' in data.keys():
			@self.error(data['error'])
			def route(*args, **kwargs):
				return(getattr(self.route_list, page)(self, *args, **kwargs))

		else:
			return()

		return(route)

	def build_routes(self):
		'''Build routes.'''

		# Will contain the route functions
		route_functions = {}

		# Build views
		for page in self.route_list.data.keys():
			# Retrieve route function
			route_functions[page] = self.route_builder(page,
				self.route_list.data[page])

# END ==========================================================================

################################################################################



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# ------------------------------------------------------------------------------
# 
# Author: Gabriele Girelli
# Email: gigi.ga90@gmail.com
# Version: 0.0.1
# Description: Probe-design Route manager
# 
# ------------------------------------------------------------------------------



# DEPENDENCIES =================================================================

import bottle as bot
import configparser
import datetime
import hashlib
import os
import pandas as pd
import shlex
import time
import zipfile

import ifpd as fp
from ifpd.sections import routes
from ifpd.sections.probe_design.query import Query

# ==============================================================================

def zipFile(path, ziph, root = None):
	'''Zips a file.
	Args:
		path (string): file path.
		ziph (Zipfile): zipfile handler.
	Returns:
		None.
	'''
	assert os.path.isfile(path), 'file expected.'
	oldRoot = ""
	new_path = path
	if not type(None) == type(root):
		oldRoot = os.path.commonpath([root, path])
		if 0 != len(oldRoot):
			if new_path.startswith(oldRoot):
				new_path = new_path[len(oldRoot):]
	ziph.write(path, new_path)

def zipDir(path, ziph, mainRoot = None):
	'''Zips a directory. Based on https://goo.gl/GeCPVm.
	Args:
		path (string): directory path.
		ziph (Zipfile): zipfile handler.
	Returns:
		None.
	'''

	assert os.path.isdir(path), 'folder expected.'
	for root, dirs, files in os.walk(path):
		for file in files:
			zipFile(os.path.join(root, file), ziph, mainRoot)

class Routes(routes.Routes):
	'''Routes class.'''
	
	# Empty routes dictionary
	data = {}
	zipDirName = 'zips'

	def __init__(self):
		'''
		Create new routes here using the add_route method.
		'''

		# Set default Routes
		super(Routes, self).__init__()

		# Static files ---------------------------------------------------------

		route = '/q/<query_id>/download/'
		self.add_route('query_download', 'route', route)

		dname = ('<dname:re:(images|documents)>',)
		route = '/q/<query_id>/c/<candidate_id>/%s/<path>' % dname
		self.add_route('candidate_static_file', 'route', route)
		route = '/q/<query_id>/c/<candidate_id>/documents/'
		route += '<path:re:.*>/download/'
		self.add_route('candidate_static_file_download', 'route', route)
		route = '/q/<query_id>/c/<candidate_id>/download/'
		self.add_route('candidate_download', 'route', route)

		dname = ('<dname:re:(images|documents)>',)
		route = '/q/<query_id>/cs/<candidate_id>/%s/<path>' % dname
		self.add_route('candidate_set_static_file', 'route', route)
		route = '/q/<query_id>/cs/<candidate_id>/documents/'
		route += '<path:re:.*>/download/'
		self.add_route('candidate_set_static_file_download', 'route', route)
		route = '/q/<query_id>/cs/<candidate_id>/download/'
		self.add_route('candidate_set_download', 'route', route)

		dname = ('<dname:re:(images|documents)>',)
		route = '/q/<query_id>/cs/<candidate_id>/p/<probe_id>/%s/<path>' % dname
		self.add_route('candidate_set_probe_static_file', 'route', route)
		route = '/q/<query_id>/cs/<candidate_id>/p/<probe_id>/documents/'
		route += '<path:re:.*>/download/'
		self.add_route('candidate_set_probe_static_file_download',
			'route', route)
		route = '/q/<query_id>/cs/<candidate_id>/p/<probe_id>/download/'
		self.add_route('candidate_set_probe_download', 'route', route)

		# Pages ----------------------------------------------------------------

		self.add_route('home', 'route', '/')
		self.add_route('home', 'view', 'home.tpl')

		self.add_route('query', 'route', '/q/<query_id>')
		self.add_route('query', 'view', 'query.tpl')

		uri = '/q/<query_id>/c/<candidate_id>'
		self.add_route('candidate_probe', 'route', uri)
		self.add_route('candidate_probe', 'view', 'candidate_probe.tpl')

		uri = '/q/<query_id>/cs/<candidate_id>'
		self.add_route('candidate_set', 'route', uri)
		self.add_route('candidate_set', 'view', 'candidate_set.tpl')

		uri = '/q/<query_id>/cs/<candidate_id>/p/<probe_id>'
		self.add_route('candidate_set_probe', 'route', uri)
		self.add_route('candidate_set_probe', 'view', 'candidate_set_probe.tpl')

		# Forms ----------------------------------------------------------------

		self.add_route('single_query', 'post', '/single_query')

		self.add_route('spotting_query', 'post', '/spotting_query')

		self.add_route('single_queries', 'post', '/single_queries')

		self.add_route('hide_alert', 'post', '/hide_alert')

		# Errors ---------------------------------------------------------------

		self.add_route('error404', 'error', 404)
		self.add_route('error500', 'error', 500)

		# AJAX requests --------------------------------------------------------

		self.add_route('list_chromosomes', 'get', '/listChr/<dbDir>')
		self.add_route('queueStatus', 'get', '/queueStatus')

		return

	def mkZipDir(routes, self):
		zipDirPath = os.path.join(self.static_path, 'query', routes.zipDirName)
		if not os.path.isdir(zipDirPath):
			os.mkdir(zipDirPath)

	def zipQuery(routes, self, query_id):
		zipDirPath = os.path.join(self.static_path, 'query', routes.zipDirName)
		zipPath = os.path.join(zipDirPath, f'{query_id}.zip')
		zipf = zipfile.ZipFile(zipPath, 'w', zipfile.ZIP_DEFLATED)
		zipDir(os.path.join(self.static_path, 'query', query_id), zipf,
			os.path.join(self.static_path, 'query'))
		zipFile(os.path.join(self.static_path, 'query',
			f'{query_id}.config'), zipf,
			os.path.join(self.static_path, 'query'))
		zipf.close()

	def zipCandidate(routes, self, query_id, candidate_id):
		zipDirPath = os.path.join(self.static_path, 'query', routes.zipDirName)
		zipPath = os.path.join(zipDirPath,
			f'{query_id}.candidate_{candidate_id}.zip')
		zipf = zipfile.ZipFile(zipPath, 'w', zipfile.ZIP_DEFLATED)
		zipDir(
			os.path.join(self.static_path, 'query',
				query_id, f'candidate_{candidate_id}'),
			zipf,
			os.path.join(self.static_path, 'query', query_id,
				f'candidate_{candidate_id}'))
		zipf.close()

	def zipCandidateSet(routes, self, query_id, candidate_id):
		zipDirPath = os.path.join(self.static_path, 'query', routes.zipDirName)
		zipPath = os.path.join(zipDirPath,
			f'{query_id}.probe_set_{candidate_id}.zip')
		zipf = zipfile.ZipFile(zipPath, 'w', zipfile.ZIP_DEFLATED)
		zipDir(
			os.path.join(self.static_path, 'query',
				query_id, f'probe_set_{candidate_id}'),
			zipf,
			os.path.join(self.static_path, 'query', query_id,
				f'probe_set_{candidate_id}'))
		zipf.close()

	def zipCandidateSetProbe(routes, self, query_id, candidate_id, probe_id):
		zipDirPath = os.path.join(self.static_path, 'query', routes.zipDirName)
		zipPath = os.path.join(zipDirPath,
			f'{query_id}.probe_set_{candidate_id}.probe_{probe_id}.zip')
		zipf = zipfile.ZipFile(zipPath, 'w', zipfile.ZIP_DEFLATED)
		zipDir(
			os.path.join(self.static_path, 'query',
				query_id, f'probe_set_{candidate_id}', f'probe_{probe_id}'),
			zipf,
			os.path.join(self.static_path, 'query', query_id,
				f'probe_set_{candidate_id}', f'probe_{probe_id}'))
		zipf.close()

	# Static files -------------------------------------------------------------

	def candidate_static_file(routes, self,
		query_id, candidate_id, dname, path):
		'''Access candidate static files.

		Args:
			self (App): ProbeDesigner.App instance.
			query_id (string): query folder name.
			candidate_id (string): candidate folder name.
			dname (string): file type.
			path (string): file name.
		'''
		ipath = '%s/query/%s/candidate_%s/' % (
			self.static_path, query_id, candidate_id)
		return(bot.static_file(path, ipath))

	def candidate_static_file_download(routes, self,
		query_id, candidate_id, path):
		'''Download candidate static files.

		Args:
			self (App): ProbeDesigner.App instance.
			query_id (string): query folder name.
			candidate_id (string): candidate folder name.
			dname (string): file type.
			path (string): file name.
		'''
		ipath = '%s/query/%s/candidate_%s/' % (
			self.static_path, query_id, candidate_id)
		outname = '%s.%s' % (query_id, path)
		return(bot.static_file(path, ipath, download = outname))

	def candidate_set_static_file(routes, self,
		query_id, candidate_id, dname, path):
		'''Access candidate static files.

		Args:
			self (App): ProbeDesigner.App instance.
			query_id (string): query folder name.
			candidate_id (string): candidate folder name.
			dname (string): file type.
			path (string): file name.
		'''
		ipath = '%s/query/%s/probe_set_%s/' % (
			self.static_path, query_id, candidate_id)
		return(bot.static_file(path, ipath))

	def candidate_set_static_file_download(routes, self,
		query_id, candidate_id, path):
		'''Download candidate static files.

		Args:
			self (App): ProbeDesigner.App instance.
			query_id (string): query folder name.
			candidate_id (string): candidate folder name.
			dname (string): file type.
			path (string): file name.
		'''
		ipath = '%s/query/%s/probe_set_%s/' % (
			self.static_path, query_id, candidate_id)
		outname = '%s.%s' % (query_id, path)
		return(bot.static_file(path, ipath, download = outname))

	def candidate_set_probe_static_file(routes, self,
		query_id, candidate_id, probe_id, dname, path):
		'''Access candidate static files.

		Args:
			self (App): ProbeDesigner.App instance.
			query_id (string): query folder name.
			candidate_id (string): candidate folder name.
			dname (string): file type.
			path (string): file name.
		'''
		ipath = '%s/query/%s/probe_set_%s/probe_%s/' % (
			self.static_path, query_id, candidate_id, probe_id)
		return(bot.static_file(path, ipath))

	def candidate_set_probe_static_file_download(routes, self,
		query_id, candidate_id, probe_id, path):
		'''Download candidate static files.

		Args:
			self (App): ProbeDesigner.App instance.
			query_id (string): query folder name.
			candidate_id (string): candidate folder name.
			dname (string): file type.
			path (string): file name.
		'''
		ipath = '%s/query/%s/probe_set_%s/probe_%s/' % (
			self.static_path, query_id, candidate_id, probe_id)
		outname = '%s.probe_set_%s.%s' % (query_id, candidate_id, path)
		return(bot.static_file(path, ipath, download = outname))

	def query_download(routes, self,
		query_id):
		'''Download compressed query output.

		Args:
			self (App): ProbeDesigner.App instance.
			query_id (string): query folder name.
		'''

		routes.mkZipDir(self)
		ipath = os.path.join(self.static_path, 'query', routes.zipDirName)
		fName = f'{query_id}.zip'
		outname = f'query.{fName}'

		if not os.path.isfile(os.path.join(ipath, fName)):
			routes.zipQuery(self, query_id)

		return(bot.static_file(fName, ipath, download = outname))

	def candidate_download(routes, self,
		query_id, candidate_id):
		'''Download compressed candidate output.

		Args:
			self (App): ProbeDesigner.App instance.
			query_id (string): query folder name.
			candidate_id (string): candidate folder name.
		'''

		routes.mkZipDir(self)
		ipath = os.path.join(self.static_path, 'query', routes.zipDirName)
		fName = f'{query_id}.candidate_{candidate_id}.zip'
		outname = f'query.{fName}'

		if not os.path.isfile(os.path.join(ipath, fName)):
			routes.zipCandidate(self, query_id, candidate_id)

		return bot.static_file(fName, ipath, download = outname)

	def candidate_set_download(routes, self,
		query_id, candidate_id):
		'''Download compressed candidate output.

		Args:
			self (App): ProbeDesigner.App instance.
			query_id (string): query folder name.
			candidate_id (string): candidate folder name.
		'''

		routes.mkZipDir(self)
		ipath = os.path.join(self.static_path, 'query', routes.zipDirName)
		fName = f'{query_id}.probe_set_{candidate_id}.zip'
		outname = f'query.{fName}'

		if not os.path.isfile(os.path.join(ipath, fName)):
			routes.zipCandidateSet(self, query_id, candidate_id)

		return bot.static_file(fName, ipath, download = outname)

	def candidate_set_probe_download(routes, self,
		query_id, candidate_id, probe_id):
		'''Download compressed candidate output.

		Args:
			self (App): ProbeDesigner.App instance.
			query_id (string): query folder name.
			candidate_id (string): candidate folder name.
		'''

		routes.mkZipDir(self)
		ipath = os.path.join(self.static_path, 'query', routes.zipDirName)
		fName = f'{query_id}.probe_set_{candidate_id}.probe_{probe_id}.zip'
		outname = f'query.{fName}'

		if not os.path.isfile(os.path.join(ipath, fName)):
			routes.zipCandidateSetProbe(self, query_id, candidate_id, probe_id)

		return bot.static_file(fName, ipath, download = outname)

	# Pages --------------------------------------------------------------------

	def home(routes, self):
		'''Home-page.

		Args:
			self (App): ProbeDesigner.App instance.
		'''

		d = self.vd

		d['title'] = self.tprefix + 'Home'
		d['custom_stylesheets'] = ['home.css', 'style.css']
		d['custom_root_stylesheets'] = []

		
		configPathList = [os.path.join(self.static_path, 'db', p, '.config')
			for p in next(os.walk(os.path.join(self.static_path, 'db')))[1]]
		configPathList.sort()
		d['dbdata'] = []
		for configPath in configPathList:
			with open(configPath, 'r') as IH:
				parser = configparser.ConfigParser()
				parser.read_string("".join(IH.readlines()))
				d['dbdata'].append(parser)
		d['dblist'] = {}
		for p in d['dbdata']:
			dbDir = os.path.basename(p['SOURCE']['outdirectory'])
			dbName = p['DATABASE']['name']
			d['dblist'][dbName] = dbDir
		
		return(d)

	def query(routes, self, query_id):
		'''Query output page.

		Args:
			self (App): ProbeDesigner.App instance.
			query_id (string): query folder name.
		'''

		d = self.vd

		d['title'] = '%s Query: %s' % (self.tprefix, query_id)
		d['custom_stylesheets'] = ['query.css', 'style.css']
		d['custom_root_stylesheets'] = []

		d['query'] = Query(query_id, self.qpath).data
		d['queryRoot'] = self.qpath

		if 'done' == d['query']['status']:
			if 'single' == d['query']['type']:
				d['query']['candidate_table'] = pd.read_csv(
					os.path.join(self.qpath, query_id, "candidates.tsv"),
					"\t")
			elif 'spotting' == d['query']['type']:
				d['query']['candidate_table'] = pd.read_csv(
					os.path.join(self.qpath, query_id, "set_candidates.tsv"),
					"\t")

		d['queryTimeout'] = 24*60*60 # 1 day timeout

		d['admin_email'] = self.admin_email

		return(d)

	def candidate_probe(routes, self, query_id, candidate_id):
		'''Candidate output page.

		Args:
			self (App): ProbeDesigner.App instance.
			query_id (string): query folder name.
			candidate_id (string): candidate folder name.
		'''

		d = self.vd

		d['title'] = '%s Query: %s' % (self.tprefix, query_id)
		d['custom_stylesheets'] = ['query.css', 'style.css']
		d['custom_root_stylesheets'] = []
		
		d['query'] = Query(query_id, self.qpath).data
		d['queryRoot'] = self.qpath

		d['candidate'] = {'id' : candidate_id}
		configPath = os.path.join(self.qpath, query_id,
			f"candidate_{candidate_id}", f"candidate_{candidate_id}.config")
		with open(configPath, 'r') as IH:
			config = configparser.ConfigParser()
			config.read_string("".join(IH.readlines()))
			d['candidate'].update(config['REGION'].items())
			d['candidate'].update(config['PROBE'].items())
			d['candidate'].update(config['FEATURES'].items())

		return(d)

	def candidate_set(routes, self, query_id, candidate_id):
		'''Candidate output page.

		Args:
			self (App): ProbeDesigner.App instance.
			query_id (string): query folder name.
			candidate_id (string): candidate folder name.
		'''

		d = self.vd

		d['title'] = '%s Query: %s' % (self.tprefix, query_id)
		d['custom_stylesheets'] = ['query.css', 'style.css']
		d['custom_root_stylesheets'] = []
		
		d['query'] = Query(query_id, self.qpath).data
		d['queryRoot'] = self.qpath
		d['query']['candidate_table'] = pd.read_csv(
			os.path.join(self.qpath, query_id, "set_candidates.tsv"), "\t")

		d['candidate'] = {'id' : candidate_id}

		return(d)

	def candidate_set_probe(routes, self, query_id, candidate_id, probe_id):
		'''Candidate output page.

		Args:
			self (App): ProbeDesigner.App instance.
			query_id (string): query folder name.
			candidate_id (string): candidate folder name.
		'''

		d = self.vd

		d['title'] = '%s Query: %s' % (self.tprefix, query_id)
		d['custom_stylesheets'] = ['query.css', 'style.css']
		d['custom_root_stylesheets'] = []
		
		d['query'] = Query(query_id, self.qpath).data
		d['queryRoot'] = self.qpath
		d['query']['candidate_table'] = pd.read_csv(
			os.path.join(self.qpath, query_id, "set_candidates.tsv"), "\t")

		d['candidate'] = {'id' : candidate_id}

		d['probe'] = {'id' : probe_id}
		configPath = os.path.join(self.qpath, query_id,
			f"probe_set_{candidate_id}", f"probe_{probe_id}",
			f"probe_{probe_id}.config")
		with open(configPath, 'r') as IH:
			config = configparser.ConfigParser()
			config.read_string("".join(IH.readlines()))
			d['probe'].update(config['REGION'].items())
			d['probe'].update(config['PROBE'].items())
			d['probe'].update(config['FEATURES'].items())

		return(d)

	# Form reception -----------------------------------------------------------


	def hide_alert(routes, self):
		'''Hide bookmark alert for a specific query.

		Args:
			self (App): ProbeDesigner.App instance.
		'''

		formData = bot.request.forms

		query = Query(formData.query_id, self.qpath).data
		config = configparser.ConfigParser()
		configPath = os.path.join(self.qpath, f'{formData.query_id}.config')
		with open(configPath, 'r') as IH:
			config.read_string("".join(IH.readlines()))
		config['GENERAL']['hidden_bookmark_alter'] = "True"
		with open(configPath, 'w+') as OH:
			config.write(OH)

		return 'Done'

	def single_query(routes, self):
		'''Single probe query form reception route.

		Args:
			self (App): ProbeDesigner.App instance.
		'''

		formData = bot.request.forms
		queriedRegion = f'{formData.chromosome}:{formData.start},{formData.end}'
		query_id = f'{queriedRegion}:{time.time()}'
		encoder = hashlib.sha256()
		encoder.update(bytes(query_id, "utf-8"))
		query_id = encoder.hexdigest()

		dbPath = f'{self.static_path}/db/{formData.database}'
		oligoDB = fp.query.OligoDatabase(dbPath, hasNetwork = False)
		min_dist = oligoDB.get_oligo_min_dist()

		cmd = ['ifpd_query_probe']
		cmd.extend([shlex.quote(queriedRegion), shlex.quote(dbPath)])
		cmd.extend([shlex.quote(f'{self.static_path}/query/{query_id}')])
		cmd.extend(['--order', shlex.quote(formData.f1),
			shlex.quote(formData.f2), shlex.quote(formData.f3)])
		cmd.extend(['--filter-thr', shlex.quote(formData.f1_threshold)])
		cmd.extend(['--n-oligo', shlex.quote(formData.n_oligo)])
		cmd.extend(['--max-probes', shlex.quote(formData.max_probes)])
		cmd.extend(['--min-d', shlex.quote(f'{min_dist}')])

		config = configparser.ConfigParser()
		timestamp = time.time()
		config['GENERAL'] = {
			'name' : formData.name,
			'description' : formData.description,
			'type' : 'single',
			'cmd' : " ".join(cmd),
			'status' : 'queued'
		}
		config['WHEN'] = {
			'time' : timestamp,
			'isotime' : datetime.datetime.fromtimestamp(timestamp).isoformat()
		}
		config['WHERE'] = {
			'db' : formData.database,
			'region' : queriedRegion
		}
		config['WHAT'] = {
			'n_oligo' : formData.n_oligo,
			'threshold' : formData.f1_threshold,
			'max_probes' : formData.max_probes
		}
		config['HOW'] = {
			'f1' : formData.f1,
			'f2' : formData.f2,
			'f3' : formData.f3
		}
		configPath = os.path.join(self.static_path, 'query',
			f'{query_id}.config')
		with open(configPath, 'w+') as OH:
			config.write(OH)

		self.queue.put(cmd)

		bot.response.status = 303
		bot.response.set_header('Location',
			f'{self.root_uri}{self.app_uri}q/{query_id}')

		return('Query received.')

	def spotting_query(routes, self):
		'''Multi probe query form reception route.

		Args:
			self (App): ProbeDesigner.App instance.
		'''

		formData = bot.request.forms
		queriedRegion = f'{formData.multi_chromosome}:'
		queriedRegion += f'{formData.multi_start},{formData.multi_end}'
		query_id = f'{queriedRegion}:{time.time()}'
		encoder = hashlib.sha256()
		encoder.update(bytes(query_id, "utf-8"))
		query_id = encoder.hexdigest()

		dbPath = f'{self.static_path}/db/{formData.multi_database}'
		oligoDB = fp.query.OligoDatabase(dbPath, hasNetwork = False)
		min_dist = oligoDB.get_oligo_min_dist()

		cmd = ['ifpd_query_set',
			shlex.quote(queriedRegion),
			shlex.quote(formData.multi_n_probes),
			shlex.quote(dbPath),
			shlex.quote(f'{self.static_path}/query/{query_id}'),
			'--order',
				shlex.quote(formData.f1),
				shlex.quote(formData.f2),
				shlex.quote(formData.f3),
			'--filter-thr', shlex.quote(formData.multi_f1_threshold),
			'--n-oligo', shlex.quote(formData.multi_n_oligo),
			'--min-d', shlex.quote(f'{min_dist}'),
			'--window-shift', shlex.quote(f'{formData.multi_win_shift}')]

		config = configparser.ConfigParser()
		timestamp = time.time()
		config['GENERAL'] = {
			'name' : formData.multi_name,
			'description' : formData.multi_description,
			'type' : 'spotting',
			'cmd' : " ".join(cmd),
			'status' : 'queued'
		}
		config['WHEN'] = {
			'time' : timestamp,
			'isotime' : datetime.datetime.fromtimestamp(timestamp).isoformat()
		}
		config['WHERE'] = {
			'db' : formData.multi_database,
			'region' : queriedRegion
		}
		config['WHAT'] = {
			'n_oligo' : formData.multi_n_oligo,
			'threshold' : formData.multi_f1_threshold,
			'n_probes' : formData.multi_n_probes,
			'window_shift' : formData.multi_win_shift
		}
		config['HOW'] = {
			'f1' : formData.f1,
			'f2' : formData.f2,
			'f3' : formData.f3
		}
		configPath = os.path.join(self.static_path, 'query',
			f'{query_id}.config')
		with open(configPath, 'w+') as OH:
			config.write(OH)

		self.queue.put(cmd)

		bot.response.status = 303
		bot.response.set_header('Location',
			f'{self.root_uri}{self.app_uri}q/{query_id}')

		# Output
		return('Query received.')

	def single_queries(routes, self):
		'''Single probe queries reception route.

		Args:
			self (App): ProbeDesigner.App instance.
		'''

		# Read uploaded file
		data = bot.request.files.data
		raw = [row.decode('utf-8') for row in data.file.readlines()]

		# Submit query
		for row in raw:
			args = row.strip().split('\t')

			# Build query command line
			cmd = ['fprode_dbquery']
			cmd.extend([Query.get_next_id(self.qpath, self.queue)])
			cmd.extend([shlex.quote(args[0])])
			cmd.extend([shlex.quote(e) for e in args[3:6]])
			cmd.extend([shlex.quote('%s/db/%s' % (self.static_path, args[2]))])
			cmd.extend(['--n_oligo', shlex.quote(args[6])])
			cmd.extend(['--f1_thr', shlex.quote(args[7])])
			cmd.extend(['--max_probes', shlex.quote(args[8])])
			cmd.extend(['--feat_order', shlex.quote(args[9])])
			cmd.extend(['--outdir', '%s/query/' % self.static_path])
			cmd.extend(['--description', shlex.quote(args[1])])

			# Add query to the queue
			self.queue.put(cmd)

		# Redirect
		bot.response.status = 303
		bot.response.set_header('Location',
			"%s%s" % (self.root_uri, self.app_uri))

		# Output
		return('Query received.')

	# Error --------------------------------------------------------------------

	# AJAX Requests ------------------------------------------------------------

	def list_chromosomes(routes, self, dbDir):
		dbPath = os.path.join(self.static_path, 'db', dbDir)
		chrList = [x for x in os.listdir(dbPath) if not os.path.isdir(x)]
		chrList = [x for x in chrList if not x in ['.log', '.config']]
		if 0 == len(chrList): return '{"chrList":[]}'
		chrList.sort()
		return '{"chrList":["%s"]}' % '","'.join(chrList[::-1])

	def queueStatus(routes, self):
		taskList = []
		for task in self.queue.queue:
			if 'ifpd_query_set' == task[0]:
				outdir_id = 4
			elif 'ifpd_query_probe' == task[0]:
				outdir_id = 3
			query_id = os.path.basename(task[outdir_id])
			data = Query(query_id, self.qpath).data
			taskList.append(os.path.basename(task[1]) + f' @{data["isotime"]}')
		if 0 == len(taskList): return '{"queue":[]}'
		return '{"queue": ["%s"]}' % '", "'.join(taskList)

# END ==========================================================================

################################################################################



# -*- coding: utf-8 -*-

'''
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
@description: web-services and related functions.
'''



# ==============================================================================

import urllib.request, urllib.error
import xml.etree.ElementTree

UCSC_DAS_URI = 'http://genome.ucsc.edu/cgi-bin/das/'

def internet_on():
    '''Check internet connection status.
    From: https://stackoverflow.com/a/3764660'''
    try:
        urllib.request.urlopen('http://216.58.192.142', timeout = 1)
        return True
    except urllib.error.URLError as e: 
        return False

def get_webpage_content(uri):
    '''Get the content of a web-page.'''
    try:
        page = urllib.request.urlopen(uri)
        pageContent = page.read()
        page.close()
    except urllib.error.URLError as e:
        raise
    return pageContent

def assert_UCSC_DSN(dsn):
    '''Runs assert checks on a UCSC DAS DSN (XML format only!).'''
    assert 0 < len(dsn), "no databases found."
    attrib = dsn[0].attrib
    assert "id" in attrib.keys(), "database has no 'id' key."
    assert "version" in attrib.keys(), "database has no 'version' key."

def print_UCSC_DSN(dsn):
    '''Prints info on a UCSC DAS DSN (XML format only!).'''
    attrib = dsn[0].attrib
    print(f'"{attrib["id"]}" (v{attrib["version"]}) : {dsn[0].text}')

def list_UCSC_reference_genomes(verbose = False, UCSC_DAS_URI = UCSC_DAS_URI):
    '''Retrieves list of UCSC DAS reference genome IDs.
    Use verbose to print a readable list.'''

    assert internet_on(), "cannot connect to the internet."

    dsnXMLdata = get_webpage_content(f'{UCSC_DAS_URI}/dsn')

    databases = set()
    for dsn in xml.etree.ElementTree.fromstring(dsnXMLdata):
        assert_UCSC_DSN(dsn)
        if verbose:
            print_UCSC_DSN(dsn)
        databases.add(dsn[0].attrib["id"])

    return(databases)

def get_sequence_from_UCSC(region, genome, UCSC_DAS_URI = UCSC_DAS_URI):
    '''Retrieve the sequence of a certain reference genome region from UCSC DAS
    server. As the input region definition follows the bed format, the chromEnd
    position is not included.'''
    chrom, chromStart, chromEnd = region
    base_uri = f'{UCSC_DAS_URI}/{genome}/dna'
    uri_query = f'?segment={chrom}:{chromStart},{chromEnd-1}'
    seqXMLdata = get_webpage_content(base_uri+uri_query)

    seq = xml.etree.ElementTree.fromstring(seqXMLdata)[0][0].text
    seq = seq.replace('\n', '').replace('\r', '').replace(' ', '')
    return seq.upper()

def get_segment_size_from_UCSC(genome, chrom, UCSC_DAS_URI = UCSC_DAS_URI):
    if chrom.startswith('chr'):
        chrom = chrom[3:]

    base_uri = f'{UCSC_DAS_URI}/{genome}/entry_points'
    chromXMLdata = get_webpage_content(base_uri)

    chromList = xml.etree.ElementTree.fromstring(chromXMLdata)[0]
    chromIDs = [chrom.attrib['id'] for chrom in chromList]
    
    return int(chromList[chromIDs.index(chrom)].attrib['stop'])

def check_reference_genome(refGenome, UCSC_DAS_URI = UCSC_DAS_URI):
    refGenomeList = list_UCSC_reference_genomes(
      UCSC_DAS_URI = UCSC_DAS_URI)
    return refGenome in refGenomeList

def get_chromosome_size(chrom, refGenome,
    UCSC_DAS_URI = UCSC_DAS_URI):
    return get_segment_size_from_UCSC(refGenome, chrom,
        UCSC_DAS_URI = UCSC_DAS_URI)

def check_chromosome_size(chrom, chromSize, refGenome,
    UCSC_DAS_URI = UCSC_DAS_URI):
    return chromSize <= get_chromosome_size(chrom, refGenome, UCSC_DAS_URI)

def check_sequence(region, sequence, refGenome, UCSC_DAS_URI = UCSC_DAS_URI):
    '''Connects to UCSC to check if a sequence is correct.'''
    regionSequence = get_sequence_from_UCSC(
      region, refGenome, UCSC_DAS_URI)
    assert_msg = f'\nSequence: {sequence}'
    assert_msg += f'\nUCSC seq: {regionSequence}'
    return (sequence == regionSequence, assert_msg)

# END ==========================================================================

################################################################################



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# Get the long description from the README file
with open('README.md', "r") as f:
	long_description = f.read()

setup(name='ifpd',
	version='2.0.1.post1',
	description='''An iFISH probe design pipeline, with web interface included.''',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url='https://github.com/ggirelli/iFISH-Probe-Design',
	author='Gabriele Girelli',
	author_email='gabriele.girelli@scilifelab.se',
	license='MIT',
	classifiers=[
		'Development Status :: 4 - Beta',
		'Intended Audience :: Science/Research',
		'Topic :: Scientific/Engineering :: Bio-Informatics',
		'License :: OSI Approved :: MIT License',
		'Programming Language :: Python :: 3 :: Only',
	],
	keywords='biology cell DNA RNA FISH fluorescence hybridization bioimaging genome',
	packages=["ifpd", "ifpd.sections", "ifpd.sections.probe_design"],
	install_requires=[
		"bottle>=0.12.13",
		'ggc>=0.0.3',
		'joblib>=0.11',
		"matplotlib>=3.0.0",
		"numpy>=1.14.2",
		"pandas>=0.22.0",
		"paste>=2.0.3",
		"scipy>=1.0.0"
	],
	scripts=[
		"bin/ifpd_mkdb",
		"bin/ifpd_dbchk",
		"bin/ifpd_query_probe",
		"bin/ifpd_query_set",
		"bin/ifpd_serve"
	],
	package_data={
		'ifpd': [
			'interface/css/bootstrap.css',
			'interface/css/*',
			'interface/fonts/*',
			'interface/images/*',
			'interface/js/*',
			'interface/views/*'
		],
		'ifpd.sections.probe_design': [
			'css/*', 'documents/*', 'views/*']
	},
	test_suite="nose.collector",
	tests_require=["nose"],
)


#!/usr/bin/python3
# -*- coding: utf-8 -*-

# ------------------------------------------------------------------------------
# 
# Author: Gabriele Girelli
# Email: gigi.ga90@gmail.com
# Version: 0.0.1
# Description: Probe-design Query manager
# 
# TODO:
# 	- Split Query.get_data in smaller functions.
# 
# ------------------------------------------------------------------------------



# DEPENDENCIES =================================================================

import configparser
import numpy as np
import os
import pandas as pd

# CLASSES ======================================================================

class Query(object):
	'''Single query class.'''

	def __init__(self, query_id, query_root):
		super(Query, self).__init__()

		assert_msg = f'query "{query_id}" not found.'
		assert Query.exists(query_id, query_root), assert_msg

		self.data = {}
		self.data['id'] = query_id
		with open(os.path.join(query_root, f'{query_id}.config'), 'r') as IH:
			config = configparser.ConfigParser()
			config.read_string("".join(IH.readlines()))
			self.data.update(config['GENERAL'].items())
			self.data.update(config['WHEN'].items())
			self.data.update(config['WHERE'].items())
			self.data.update(config['WHAT'].items())
			self.data.update(config['HOW'].items())

	@staticmethod
	def exists(query_id, query_root):
		'''Check if a query exists.'''
		#dirExists = os.path.isdir(os.path.join(query_root, query_id))
		configExists = os.path.isfile(
			os.path.join(query_root, f'{query_id}.config'))
		return configExists# and dirExists

# END ==========================================================================

################################################################################
